#include "config.h"

#include "core/swai-registry.h"

#include <glib.h>
#include <gio/gio.h>

typedef struct _SWAIRegistry SWAIRegistry;

struct _SWAIRegistry {
    GHashTable *title_mappings;
    GHashTable *wm_class_mappings;
    GDBusConnection *connection;
    guint registration_id;
};

static SWAIRegistry *swai_registry = NULL;

#define SWAI_REGISTRY_DBUS_SERVICE "org.gnome.Mutter.SwaiRegistry"
#define SWAI_REGISTRY_DBUS_PATH "/org/gnome/Mutter/SwaiRegistry"

static const gchar introspection_xml[] =
  "<node>"
  "  <interface name=\"org.gnome.Mutter.SwaiRegistry\">"
  "    <method name=\"RegisterTitle\">"
  "      <arg name=\"internal_title\" type=\"s\" direction=\"in\"/>"
  "      <arg name=\"display_title\" type=\"s\" direction=\"in\"/>"
  "      <arg name=\"wm_class\" type=\"s\" direction=\"in\"/>"
  "    </method>"
  "    <method name=\"UnregisterTitle\">"
  "      <arg name=\"internal_title\" type=\"s\" direction=\"in\"/>"
  "    </method>"
  "    <signal name=\"title_changed\">"
  "      <arg name=\"internal_title\" type=\"s\" direction=\"out\"/>"
  "      <arg name=\"display_title\" type=\"s\" direction=\"out\"/>"
  "    </signal>"
  "  </interface>"
  "</node>";

static GDBusNodeInfo *introspection_data = NULL;

static void on_bus_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data);
static void on_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data);
static void on_name_lost(GDBusConnection *connection, const gchar *name, gpointer user_data);

static guint name_owner_id = 0;

// Helper function to remove window ID from the end of wm_class
static char *
strip_window_id_from_wm_class (const char *wm_class)
{
    if (!wm_class)
        return NULL;
    
    // Find the last dot in the string
    const char *last_dot = strrchr(wm_class, '.');
    if (!last_dot)
        return g_strdup(wm_class);
    
    // Check if everything after the last dot is a number
    const char *potential_id = last_dot + 1;
    if (*potential_id == '\0')
        return g_strdup(wm_class);
    
    // Check if it's all digits
    for (const char *c = potential_id; *c; c++) {
        if (!g_ascii_isdigit(*c))
            return g_strdup(wm_class);
    }
    
    // If we get here, everything after the last dot is a number
    // Return a copy without the window ID
    return g_strndup(wm_class, last_dot - wm_class);
}

static void swai_registry_method_call(
    GDBusConnection *connection,
    const gchar *sender,
    const gchar *object_path,
    const gchar *interface_name,
    const gchar *method_name,
    GVariant *parameters,
    GDBusMethodInvocation *invocation,
    gpointer user_data)
{
    if (g_strcmp0(method_name, "RegisterTitle") == 0) {
        char *internal_title, *display_title, *wm_class;
        g_variant_get(parameters, "(sss)", &internal_title, &display_title, &wm_class);
        
        // Register the title mapping
        g_hash_table_replace(swai_registry->title_mappings, g_strdup(internal_title), g_strdup(display_title));
        
        // Register the wm_class mapping with window ID stripped
        char *stripped_wm_class = strip_window_id_from_wm_class(wm_class);
        if (stripped_wm_class) {
            g_hash_table_replace(swai_registry->wm_class_mappings, g_strdup(wm_class), stripped_wm_class);
        }
        
        g_dbus_connection_emit_signal(
            connection,
            NULL,
            SWAI_REGISTRY_DBUS_PATH,
            "org.gnome.Mutter.SwaiRegistry",
            "title_changed",
            g_variant_new("(ss)", internal_title, display_title),
            NULL);
        g_dbus_method_invocation_return_value(invocation, NULL);
        g_free(internal_title);
        g_free(display_title);
        g_free(wm_class);
    } else if (g_strcmp0(method_name, "UnregisterTitle") == 0) {
        char *internal_title;
        g_variant_get(parameters, "(s)", &internal_title);
        g_hash_table_remove(swai_registry->title_mappings, internal_title);
        // Note: We don't remove from wm_class_mappings here since multiple titles 
        // might share the same base wm_class
        g_dbus_connection_emit_signal(
            connection,
            NULL,
            SWAI_REGISTRY_DBUS_PATH,
            "org.gnome.Mutter.SwaiRegistry",
            "title_changed",
            g_variant_new("(ss)", internal_title, ""),
            NULL);
        g_dbus_method_invocation_return_value(invocation, NULL);
        g_free(internal_title);
    } else {
        g_dbus_method_invocation_return_error(invocation, G_DBUS_ERROR, G_DBUS_ERROR_UNKNOWN_METHOD, "Unknown method: %s", method_name);
    }
}

static GDBusInterfaceVTable interface_vtable = {
    .method_call = swai_registry_method_call,
    .get_property = NULL,
    .set_property = NULL
};

static void on_bus_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data) {
    guint registration_id;
    registration_id = g_dbus_connection_register_object(
        connection,
        SWAI_REGISTRY_DBUS_PATH,
        introspection_data->interfaces[0],
        &interface_vtable,
        NULL, /* user_data */
        NULL, /* user_data_free_func */
        NULL  /* GError** */
    );
    if (registration_id == 0) {
        g_warning("Failed to register SwaiRegistry D-Bus object");
    } else {
        swai_registry->connection = g_object_ref(connection);
        swai_registry->registration_id = registration_id;
    }
}

static void on_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data) {
    g_message("SwaiRegistry D-Bus name acquired: %s", name);
}

static void on_name_lost(GDBusConnection *connection, const gchar *name, gpointer user_data) {
    g_warning("SwaiRegistry D-Bus name lost: %s", name);
}

void swai_registry_init(void) {
    swai_registry = g_new0(SWAIRegistry, 1);
    swai_registry->title_mappings = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    swai_registry->wm_class_mappings = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    introspection_data = g_dbus_node_info_new_for_xml(introspection_xml, NULL);
    name_owner_id = g_bus_own_name(
        G_BUS_TYPE_SESSION,
        SWAI_REGISTRY_DBUS_SERVICE,
        G_BUS_NAME_OWNER_FLAGS_NONE,
        on_bus_acquired,
        on_name_acquired,
        on_name_lost,
        NULL,
        NULL
    );
}

void swai_registry_shutdown(void) {
    if (swai_registry) {
        if (swai_registry->registration_id) {
            g_dbus_connection_unregister_object(swai_registry->connection, swai_registry->registration_id);
            swai_registry->registration_id = 0;
        }
        g_hash_table_unref(swai_registry->title_mappings);
        g_hash_table_unref(swai_registry->wm_class_mappings);
        g_clear_object(&swai_registry->connection);
        g_free(swai_registry);
        swai_registry = NULL;
    }
    if (introspection_data) {
        g_dbus_node_info_unref(introspection_data);
        introspection_data = NULL;
    }
    if (name_owner_id) {
        g_bus_unown_name(name_owner_id);
        name_owner_id = 0;
    }
}

const char *swai_registry_get_display_title (const char *internal_title) {
    if (!swai_registry) {
        return internal_title;
    }
    const char *display_title = g_hash_table_lookup(swai_registry->title_mappings, internal_title);
    if (!display_title) {
        return internal_title;
    }
    return display_title;
}

const char *swai_registry_get_display_wm_class (const char *internal_wm_class) {
    if (!swai_registry || !internal_wm_class) {
        return internal_wm_class;
    }
    const char *display_wm_class = g_hash_table_lookup(swai_registry->wm_class_mappings, internal_wm_class);
    if (!display_wm_class) {
        return internal_wm_class;
    }
    return display_wm_class;
}